@@ -47,17 +47,18 @@ static inline int mystrcmp(char **p, const char *sample)
         return 0;
 }
 
-static void change_font_size(ASS_Renderer *render_priv, double sz)
+double ensure_font_size(ASS_Renderer *priv, double size)
 {
-    double size = sz * render_priv->font_scale;
-
     if (size < 1)
         size = 1;
-    else if (size > render_priv->height * 2)
-        size = render_priv->height * 2;
+    else if (size > priv->height * 2)
+        size = priv->height * 2;
 
-    ass_font_set_size(render_priv->state.font, size);
+    return size;
+}
 
+static void change_font_size(ASS_Renderer *render_priv, double sz)
+{
     render_priv->state.font_size = sz;
 }
 
@@ -80,14 +81,14 @@ void update_font(ASS_Renderer *render_priv)
 
     val = render_priv->state.bold;
     // 0 = normal, 1 = bold, >1 = exact weight
-    if (val == 1 || val == -1)
+    if (val == 1)
         val = 200;              // bold
     else if (val <= 0)
         val = 80;               // normal
     desc.bold = val;
 
     val = render_priv->state.italic;
-    if (val == 1 || val == -1)
+    if (val == 1)
         val = 110;              // italic
     else if (val <= 0)
         val = 0;                // normal
@@ -105,27 +106,14 @@ void update_font(ASS_Renderer *render_priv)
 
 /**
  * \brief Change border width
- * negative value resets border to style value
+ *
+ * \param render_priv renderer state object
+ * \param info glyph state object
  */
-void change_border(ASS_Renderer *render_priv, double border_x,
-                   double border_y)
+void change_border(ASS_Renderer *render_priv, double border_x, double border_y)
 {
-    int bord;
-    if (!render_priv->state.font)
-        return;
+    int bord = 64 * border_x * render_priv->border_scale;
 
-    if (border_x < 0 && border_y < 0) {
-        if (render_priv->state.style->BorderStyle == 1 ||
-            render_priv->state.style->BorderStyle == 3)
-            border_x = border_y = render_priv->state.style->Outline;
-        else
-            border_x = border_y = 1.;
-    }
-
-    render_priv->state.border_x = border_x;
-    render_priv->state.border_y = border_y;
-
-    bord = 64 * border_x * render_priv->border_scale;
     if (bord > 0 && border_x == border_y) {
         if (!render_priv->state.stroker) {
             int error;
@@ -137,11 +125,14 @@ void change_border(ASS_Renderer *render_priv, double border_x,
                         "failed to get stroker");
                 render_priv->state.stroker = 0;
             }
+            render_priv->state.stroker_radius = -1.0;
         }
-        if (render_priv->state.stroker)
+        if (render_priv->state.stroker && render_priv->state.stroker_radius != bord) {
             FT_Stroker_Set(render_priv->state.stroker, bord,
                            FT_STROKER_LINECAP_ROUND,
                            FT_STROKER_LINEJOIN_ROUND, 0);
+            render_priv->state.stroker_radius = bord;
+        }
     } else {
         FT_Stroker_Done(render_priv->state.stroker);
         render_priv->state.stroker = 0;
@@ -189,18 +180,19 @@ interpolate_alpha(long long now, long long t1, long long t2, long long t3,
 {
     unsigned a;
     double cf;
-    if (now <= t1) {
+
+    if (now < t1) {
         a = a1;
-    } else if (now >= t4) {
-        a = a3;
-    } else if (now < t2) {      // and > t1
+    } else if (now < t2) {
         cf = ((double) (now - t1)) / (t2 - t1);
         a = a1 * (1 - cf) + a2 * cf;
-    } else if (now > t3) {
+    } else if (now < t3) {
+        a = a2;
+    } else if (now < t4) {
         cf = ((double) (now - t3)) / (t4 - t3);
         a = a2 * (1 - cf) + a3 * cf;
-    } else {                    // t2 <= now <= t3
-        a = a2;
+    } else {                    // now >= t4
+        a = a3;
     }
 
     return a;
@@ -216,13 +208,9 @@ static char *parse_vector_clip(ASS_Renderer *render_priv, char *p)
     int res = 0;
     ASS_Drawing *drawing = render_priv->state.clip_drawing;
 
-    if (drawing && drawing->glyph)
-        FT_Done_Glyph((FT_Glyph) drawing->glyph);
     ass_drawing_free(drawing);
-    render_priv->state.clip_drawing = ass_drawing_new(
-        render_priv->fontconfig_priv,
-        render_priv->state.font,
-        render_priv->ftlibrary);
+    render_priv->state.clip_drawing =
+        ass_drawing_new(render_priv->library, render_priv->ftlibrary);
     drawing = render_priv->state.clip_drawing;
     skipopt('(');
     res = mystrtoi(&p, &scale);
@@ -232,7 +220,7 @@ static char *parse_vector_clip(ASS_Renderer *render_priv, char *p)
     drawing->scale = scale;
     drawing->scale_x = render_priv->font_scale_x * render_priv->font_scale;
     drawing->scale_y = render_priv->font_scale;
-    while (*p != ')' && *p != '}' && p != 0)
+    while (*p != ')' && *p != '}' && *p != 0)
         ass_drawing_add_char(drawing, *p++);
     skipopt(')');
 
@@ -244,7 +232,7 @@ static char *parse_vector_clip(ASS_Renderer *render_priv, char *p)
  * \param p string to parse
  * \param pwr multiplier for some tag effects (comes from \t tags)
  */
-static char *parse_tag(ASS_Renderer *render_priv, char *p, double pwr)
+char *parse_tag(ASS_Renderer *render_priv, char *p, double pwr)
 {
     skip_to('\\');
     skip('\\');
@@ -254,31 +242,33 @@ static char *parse_tag(ASS_Renderer *render_priv, char *p, double pwr)
     // New tags introduced in vsfilter 2.39
     if (mystrcmp(&p, "xbord")) {
         double val;
-        if (mystrtod(&p, &val))
+        if (mystrtod(&p, &val)) {
             val = render_priv->state.border_x * (1 - pwr) + val * pwr;
-        else
-            val = -1.;
-        change_border(render_priv, val, render_priv->state.border_y);
+            val = (val < 0) ? 0 : val;
+        } else
+            val = render_priv->state.style->Outline;
+        render_priv->state.border_x = val;
     } else if (mystrcmp(&p, "ybord")) {
         double val;
-        if (mystrtod(&p, &val))
+        if (mystrtod(&p, &val)) {
             val = render_priv->state.border_y * (1 - pwr) + val * pwr;
-        else
-            val = -1.;
-        change_border(render_priv, render_priv->state.border_x, val);
+            val = (val < 0) ? 0 : val;
+        } else
+            val = render_priv->state.style->Outline;
+        render_priv->state.border_y = val;
     } else if (mystrcmp(&p, "xshad")) {
         double val;
         if (mystrtod(&p, &val))
             val = render_priv->state.shadow_x * (1 - pwr) + val * pwr;
         else
-            val = 0.;
+            val = render_priv->state.style->Shadow;
         render_priv->state.shadow_x = val;
     } else if (mystrcmp(&p, "yshad")) {
         double val;
         if (mystrtod(&p, &val))
             val = render_priv->state.shadow_y * (1 - pwr) + val * pwr;
         else
-            val = 0.;
+            val = render_priv->state.style->Shadow;
         render_priv->state.shadow_y = val;
     } else if (mystrcmp(&p, "fax")) {
         double val;
@@ -298,7 +288,7 @@ static char *parse_tag(ASS_Renderer *render_priv, char *p, double pwr)
         int x0, y0, x1, y1;
         int res = 1;
         char *start = p;
-        skipopt('(');
+        skip('(');
         res &= mystrtoi(&p, &x0);
         skipopt(',');
         res &= mystrtoi(&p, &y0);
@@ -320,8 +310,7 @@ static char *parse_tag(ASS_Renderer *render_priv, char *p, double pwr)
         } else if (!render_priv->state.clip_drawing) {
             p = parse_vector_clip(render_priv, start);
             render_priv->state.clip_drawing_mode = 1;
-        } else
-            render_priv->state.clip_mode = 0;
+        }
     } else if (mystrcmp(&p, "blur")) {
         double val;
         if (mystrtod(&p, &val)) {
@@ -338,19 +327,23 @@ static char *parse_tag(ASS_Renderer *render_priv, char *p, double pwr)
         if (tp == 'x') {
             if (mystrtod(&p, &val)) {
                 val /= 100;
-                render_priv->state.scale_x =
-                    render_priv->state.scale_x * (1 - pwr) + val * pwr;
+                val = render_priv->state.scale_x * (1 - pwr) + val * pwr;
+                val = (val < 0) ? 0 : val;
             } else
-                render_priv->state.scale_x =
-                    render_priv->state.style->ScaleX;
+                val = render_priv->state.style->ScaleX;
+            render_priv->state.scale_x = val;
         } else if (tp == 'y') {
             if (mystrtod(&p, &val)) {
                 val /= 100;
-                render_priv->state.scale_y =
-                    render_priv->state.scale_y * (1 - pwr) + val * pwr;
+                val = render_priv->state.scale_y * (1 - pwr) + val * pwr;
+                val = (val < 0) ? 0 : val;
             } else
-                render_priv->state.scale_y =
-                    render_priv->state.style->ScaleY;
+                val = render_priv->state.style->ScaleY;
+            render_priv->state.scale_y = val;
+        } else {
+            --p;
+            render_priv->state.scale_x = render_priv->state.style->ScaleX;
+            render_priv->state.scale_y = render_priv->state.style->ScaleY;
         }
     } else if (mystrcmp(&p, "fsp")) {
         double val;
@@ -361,17 +354,17 @@ static char *parse_tag(ASS_Renderer *render_priv, char *p, double pwr)
             render_priv->state.hspacing = render_priv->state.style->Spacing;
     } else if (mystrcmp(&p, "fs+")) {
         double val;
-        if (mystrtod(&p, &val)) {
-            val = render_priv->state.font_size + pwr * val;
-        } else
+        mystrtod(&p, &val);
+        val = render_priv->state.font_size * (1 + pwr * val / 10);
+        if (val <= 0)
             val = render_priv->state.style->FontSize;
         if (render_priv->state.font)
             change_font_size(render_priv, val);
     } else if (mystrcmp(&p, "fs-")) {
         double val;
-        if (mystrtod(&p, &val))
-            val = render_priv->state.font_size - pwr * val;
-        else
+        mystrtod(&p, &val);
+        val = render_priv->state.font_size * (1 - pwr * val / 10);
+        if (val <= 0)
             val = render_priv->state.style->FontSize;
         if (render_priv->state.font)
             change_font_size(render_priv, val);
@@ -379,18 +372,21 @@ static char *parse_tag(ASS_Renderer *render_priv, char *p, double pwr)
         double val;
         if (mystrtod(&p, &val))
             val = render_priv->state.font_size * (1 - pwr) + val * pwr;
-        else
+        if (val <= 0)
             val = render_priv->state.style->FontSize;
         if (render_priv->state.font)
             change_font_size(render_priv, val);
     } else if (mystrcmp(&p, "bord")) {
-        double val;
+        double val, xval, yval;
         if (mystrtod(&p, &val)) {
-            if (render_priv->state.border_x == render_priv->state.border_y)
-                val = render_priv->state.border_x * (1 - pwr) + val * pwr;
+            xval = render_priv->state.border_x * (1 - pwr) + val * pwr;
+            yval = render_priv->state.border_y * (1 - pwr) + val * pwr;
+            xval = (xval < 0) ? 0 : xval;
+            yval = (yval < 0) ? 0 : yval;
         } else
-            val = -1.;          // reset to default
-        change_border(render_priv, val, val);
+            xval = yval = render_priv->state.style->Outline;
+        render_priv->state.border_x = xval;
+        render_priv->state.border_y = yval;
     } else if (mystrcmp(&p, "move")) {
         double x1, x2, y1, y2;
         long long t1, t2, delta_t, t;
@@ -404,22 +400,24 @@ static char *parse_tag(ASS_Renderer *render_priv, char *p, double pwr)
         mystrtod(&p, &x2);
         skip(',');
         mystrtod(&p, &y2);
+        t1 = t2 = 0;
         if (*p == ',') {
             skip(',');
             mystrtoll(&p, &t1);
             skip(',');
             mystrtoll(&p, &t2);
-            ass_msg(render_priv->library, MSGL_DBG2,
-                   "movement6: (%f, %f) -> (%f, %f), (%" PRId64 " .. %"
-                   PRId64 ")\n", x1, y1, x2, y2, (int64_t) t1,
-                   (int64_t) t2);
-        } else {
+            // VSFilter
+            if (t1 > t2) {
+                double tmp = t2;
+                t2 = t1;
+                t1 = tmp;
+            }
+        }
+        if (t1 <= 0 && t2 <= 0) {
             t1 = 0;
             t2 = render_priv->state.event->Duration;
-            ass_msg(render_priv->library, MSGL_DBG2,
-                   "movement: (%f, %f) -> (%f, %f)", x1, y1, x2, y2);
         }
-        skip(')');
+        skipopt(')');
         delta_t = t2 - t1;
         t = render_priv->time - render_priv->state.event->Start;
         if (t < t1)
@@ -465,7 +463,7 @@ static char *parse_tag(ASS_Renderer *render_priv, char *p, double pwr)
         char *start = p;
         char *family;
         skip_to('\\');
-        if (p > start) {
+        if (p > start && strncmp(start, "0", p - start)) {
             family = malloc(p - start + 1);
             strncpy(family, start, p - start);
             family[p - start] = '\0';
@@ -484,51 +482,51 @@ static char *parse_tag(ASS_Renderer *render_priv, char *p, double pwr)
                 change_alpha(&render_priv->state.c[i], a, pwr);
         } else {
             change_alpha(&render_priv->state.c[0],
-                         render_priv->state.style->PrimaryColour, pwr);
+                         render_priv->state.style->PrimaryColour, 1);
             change_alpha(&render_priv->state.c[1],
-                         render_priv->state.style->SecondaryColour, pwr);
+                         render_priv->state.style->SecondaryColour, 1);
             change_alpha(&render_priv->state.c[2],
-                         render_priv->state.style->OutlineColour, pwr);
+                         render_priv->state.style->OutlineColour, 1);
             change_alpha(&render_priv->state.c[3],
-                         render_priv->state.style->BackColour, pwr);
+                         render_priv->state.style->BackColour, 1);
         }
         // FIXME: simplify
     } else if (mystrcmp(&p, "an")) {
         int val;
-        if (mystrtoi(&p, &val) && val) {
-            int v = (val - 1) / 3;      // 0, 1 or 2 for vertical alignment
-            ass_msg(render_priv->library, MSGL_DBG2, "an %d", val);
-            if (v != 0)
-                v = 3 - v;
-            val = ((val - 1) % 3) + 1;  // horizontal alignment
-            val += v * 4;
-            ass_msg(render_priv->library, MSGL_DBG2, "align %d", val);
-            if ((render_priv->state.parsed_tags & PARSED_A) == 0) {
+        mystrtoi(&p, &val);
+        if ((render_priv->state.parsed_tags & PARSED_A) == 0) {
+            if (val >= 1 && val <= 9) {
+                int v = (val - 1) / 3;      // 0, 1 or 2 for vertical alignment
+                if (v != 0)
+                    v = 3 - v;
+                val = ((val - 1) % 3) + 1;  // horizontal alignment
+                val += v * 4;
                 render_priv->state.alignment = val;
-                render_priv->state.parsed_tags |= PARSED_A;
-            }
-        } else
-            render_priv->state.alignment =
-                render_priv->state.style->Alignment;
+            } else
+                render_priv->state.alignment =
+                    render_priv->state.style->Alignment;
+            render_priv->state.parsed_tags |= PARSED_A;
+        }
     } else if (mystrcmp(&p, "a")) {
         int val;
-        if (mystrtoi(&p, &val) && val) {
-            if ((render_priv->state.parsed_tags & PARSED_A) == 0) {
-                // take care of a vsfilter quirk: handle illegal \a8 like \a5
-                render_priv->state.alignment = (val == 8) ? 5 : val;
-                render_priv->state.parsed_tags |= PARSED_A;
-            }
-        } else
-            render_priv->state.alignment =
-                render_priv->state.style->Alignment;
+        mystrtoi(&p, &val);
+        if ((render_priv->state.parsed_tags & PARSED_A) == 0) {
+            if (val >= 1 && val <= 11)
+                // take care of a vsfilter quirk:
+                // handle illegal \a8 and \a4 like \a5
+                render_priv->state.alignment = ((val & 3) == 0) ? 5 : val;
+            else
+                render_priv->state.alignment =
+                    render_priv->state.style->Alignment;
+            render_priv->state.parsed_tags |= PARSED_A;
+        }
     } else if (mystrcmp(&p, "pos")) {
         double v1, v2;
         skip('(');
         mystrtod(&p, &v1);
         skip(',');
         mystrtod(&p, &v2);
-        skip(')');
-        ass_msg(render_priv->library, MSGL_DBG2, "pos(%f, %f)", v1, v2);
+        skipopt(')');
         if (render_priv->state.evt_type == EVENT_POSITIONED) {
             ass_msg(render_priv->library, MSGL_V, "Subtitle has a new \\pos "
                    "after \\move or \\pos, ignoring");
@@ -550,15 +548,15 @@ static char *parse_tag(ASS_Renderer *render_priv, char *p, double pwr)
         if (*p == ')') {
             // 2-argument version (\fad, according to specs)
             // a1 and a2 are fade-in and fade-out durations
-            t1 = 0;
-            t4 = render_priv->state.event->Duration;
+            t1 = -1;
             t2 = a1;
-            t3 = t4 - a2;
+            t3 = a2;
+            t4 = -1;
             a1 = 0xFF;
             a2 = 0;
             a3 = 0xFF;
         } else {
-            // 6-argument version (\fade)
+            // 7-argument version (\fade)
             // a1 and a2 (and a3) are opacity values
             skip(',');
             mystrtoi(&p, &a3);
@@ -571,7 +569,12 @@ static char *parse_tag(ASS_Renderer *render_priv, char *p, double pwr)
             skip(',');
             mystrtoll(&p, &t4);
         }
-        skip(')');
+        skipopt(')');
+        if (t1 == -1 && t4 == -1) {
+            t1 = 0;
+            t4 = render_priv->state.event->Duration;
+            t3 = t4 - t3;
+        }
         if ((render_priv->state.parsed_tags & PARSED_FADE) == 0) {
             render_priv->state.fade =
                 interpolate_alpha(render_priv->time -
@@ -580,13 +583,12 @@ static char *parse_tag(ASS_Renderer *render_priv, char *p, double pwr)
             render_priv->state.parsed_tags |= PARSED_FADE;
         }
     } else if (mystrcmp(&p, "org")) {
-        int v1, v2;
+        double v1, v2;
         skip('(');
-        mystrtoi(&p, &v1);
+        mystrtod(&p, &v1);
         skip(',');
-        mystrtoi(&p, &v2);
-        skip(')');
-        ass_msg(render_priv->library, MSGL_DBG2, "org(%d, %d)", v1, v2);
+        mystrtod(&p, &v2);
+        skipopt(')');
         if (!render_priv->state.have_origin) {
             render_priv->state.org_x = v1;
             render_priv->state.org_y = v2;
@@ -595,41 +597,37 @@ static char *parse_tag(ASS_Renderer *render_priv, char *p, double pwr)
         }
     } else if (mystrcmp(&p, "t")) {
         double v[3];
-        int v1, v2;
-        double v3;
+        double accel;
         int cnt;
         long long t1, t2, t, delta_t;
         double k;
         skip('(');
         for (cnt = 0; cnt < 3; ++cnt) {
-            if (*p == '\\')
+            if (!mystrtod(&p, &v[cnt]))
                 break;
-            mystrtod(&p, &v[cnt]);
             skip(',');
         }
         if (cnt == 3) {
-            v1 = v[0];
-            v2 = (v[1] < v1) ? render_priv->state.event->Duration : v[1];
-            v3 = v[2];
+            t1 = v[0];
+            t2 = v[1];
+            accel = v[2];
         } else if (cnt == 2) {
-            v1 = v[0];
-            v2 = (v[1] < v1) ? render_priv->state.event->Duration : v[1];
-            v3 = 1.;
+            t1 = v[0];
+            t2 = v[1];
+            accel = 1.;
         } else if (cnt == 1) {
-            v1 = 0;
-            v2 = render_priv->state.event->Duration;
-            v3 = v[0];
+            t1 = 0;
+            t2 = 0;
+            accel = v[0];
         } else {                // cnt == 0
-            v1 = 0;
-            v2 = render_priv->state.event->Duration;
-            v3 = 1.;
+            t1 = 0;
+            t2 = 0;
+            accel = 1.;
         }
         render_priv->state.detect_collisions = 0;
-        t1 = v1;
-        t2 = v2;
-        delta_t = v2 - v1;
-        if (v3 < 0.)
-            v3 = 0.;
+        if (t2 == 0)
+            t2 = render_priv->state.event->Duration;
+        delta_t = t2 - t1;
         t = render_priv->time - render_priv->state.event->Start;        // FIXME: move to render_context
         if (t <= t1)
             k = 0.;
@@ -637,17 +635,17 @@ static char *parse_tag(ASS_Renderer *render_priv, char *p, double pwr)
             k = 1.;
         else {
             assert(delta_t != 0.);
-            k = pow(((double) (t - t1)) / delta_t, v3);
+            k = pow(((double) (t - t1)) / delta_t, accel);
         }
-        while (*p == '\\')
+        while (*p != ')' && *p != '}' && *p != '\0')
             p = parse_tag(render_priv, p, k);   // maybe k*pwr ? no, specs forbid nested \t's
         skip_to(')');           // in case there is some unknown tag or a comment
-        skip(')');
+        skipopt(')');
     } else if (mystrcmp(&p, "clip")) {
         char *start = p;
         int x0, y0, x1, y1;
         int res = 1;
-        skipopt('(');
+        skip('(');
         res &= mystrtoi(&p, &x0);
         skipopt(',');
         res &= mystrtoi(&p, &y0);
@@ -669,19 +667,15 @@ static char *parse_tag(ASS_Renderer *render_priv, char *p, double pwr)
         } else if (!render_priv->state.clip_drawing) {
             p = parse_vector_clip(render_priv, start);
             render_priv->state.clip_drawing_mode = 0;
-        } else {
-            render_priv->state.clip_x0 = 0;
-            render_priv->state.clip_y0 = 0;
-            render_priv->state.clip_x1 = render_priv->track->PlayResX;
-            render_priv->state.clip_y1 = render_priv->track->PlayResY;
         }
     } else if (mystrcmp(&p, "c")) {
         uint32_t val;
         int hex = render_priv->track->track_type == TRACK_TYPE_ASS;
-        if (!strtocolor(render_priv->library, &p, &val, hex))
-            val = render_priv->state.style->PrimaryColour;
-        ass_msg(render_priv->library, MSGL_DBG2, "color: %X", val);
-        change_color(&render_priv->state.c[0], val, pwr);
+        if (strtocolor(render_priv->library, &p, &val, hex))
+            change_color(&render_priv->state.c[0], val, pwr);
+        else
+            change_color(&render_priv->state.c[0],
+                         render_priv->state.style->PrimaryColour, 1);
     } else if ((*p >= '1') && (*p <= '4') && (++p)
                && (mystrcmp(&p, "c") || mystrcmp(&p, "a"))) {
         char n = *(p - 2);
@@ -690,7 +684,7 @@ static char *parse_tag(ASS_Renderer *render_priv, char *p, double pwr)
         uint32_t val;
         int hex = render_priv->track->track_type == TRACK_TYPE_ASS;
         assert((n >= '1') && (n <= '4'));
-        if (!strtocolor(render_priv->library, &p, &val, hex))
+        if (!strtocolor(render_priv->library, &p, &val, hex)) {
             switch (n) {
             case '1':
                 val = render_priv->state.style->PrimaryColour;
@@ -708,6 +702,8 @@ static char *parse_tag(ASS_Renderer *render_priv, char *p, double pwr)
                 val = 0;
                 break;          // impossible due to assert; avoid compilation warning
             }
+            pwr = 1;
+        }
         switch (cmd) {
         case 'c':
             change_color(render_priv->state.c + cidx, val, pwr);
@@ -720,10 +716,19 @@ static char *parse_tag(ASS_Renderer *render_priv, char *p, double pwr)
                     n, cmd);
             break;
         }
-        ass_msg(render_priv->library, MSGL_DBG2, "single c/a at %f: %c%c = %X",
-               pwr, n, cmd, render_priv->state.c[cidx]);
     } else if (mystrcmp(&p, "r")) {
-        reset_render_context(render_priv);
+        char *start = p;
+        char *style;
+        skip_to('\\');
+        if (p > start) {
+            style = malloc(p - start + 1);
+            strncpy(style, start, p - start);
+            style[p - start] = '\0';
+            reset_render_context(render_priv,
+                    lookup_style_strict(render_priv->track, style));
+            free(style);
+        } else
+            reset_render_context(render_priv, NULL);
     } else if (mystrcmp(&p, "be")) {
         int val;
         if (mystrtoi(&p, &val)) {
@@ -734,81 +739,91 @@ static char *parse_tag(ASS_Renderer *render_priv, char *p, double pwr)
         } else
             render_priv->state.be = 0;
     } else if (mystrcmp(&p, "b")) {
-        int b;
-        if (mystrtoi(&p, &b)) {
-            if (pwr >= .5)
-                render_priv->state.bold = b;
-        } else
-            render_priv->state.bold = render_priv->state.style->Bold;
+        int val;
+        if (!mystrtoi(&p, &val) || !(val == 0 || val == 1 || val >= 100))
+            val = render_priv->state.style->Bold;
+        render_priv->state.bold = val;
         update_font(render_priv);
     } else if (mystrcmp(&p, "i")) {
-        int i;
-        if (mystrtoi(&p, &i)) {
-            if (pwr >= .5)
-                render_priv->state.italic = i;
-        } else
-            render_priv->state.italic = render_priv->state.style->Italic;
+        int val;
+        if (!mystrtoi(&p, &val) || !(val == 0 || val == 1))
+            val = render_priv->state.style->Italic;
+        render_priv->state.italic = val;
         update_font(render_priv);
     } else if (mystrcmp(&p, "kf") || mystrcmp(&p, "K")) {
-        int val = 0;
-        mystrtoi(&p, &val);
+        double val;
+        if (!mystrtod(&p, &val))
+            val = 100;
         render_priv->state.effect_type = EF_KARAOKE_KF;
         if (render_priv->state.effect_timing)
             render_priv->state.effect_skip_timing +=
                 render_priv->state.effect_timing;
         render_priv->state.effect_timing = val * 10;
     } else if (mystrcmp(&p, "ko")) {
-        int val = 0;
-        mystrtoi(&p, &val);
+        double val;
+        if (!mystrtod(&p, &val))
+            val = 100;
         render_priv->state.effect_type = EF_KARAOKE_KO;
         if (render_priv->state.effect_timing)
             render_priv->state.effect_skip_timing +=
                 render_priv->state.effect_timing;
         render_priv->state.effect_timing = val * 10;
     } else if (mystrcmp(&p, "k")) {
-        int val = 0;
-        mystrtoi(&p, &val);
+        double val;
+        if (!mystrtod(&p, &val))
+            val = 100;
         render_priv->state.effect_type = EF_KARAOKE;
         if (render_priv->state.effect_timing)
             render_priv->state.effect_skip_timing +=
                 render_priv->state.effect_timing;
         render_priv->state.effect_timing = val * 10;
     } else if (mystrcmp(&p, "shad")) {
-        double val;
+        double val, xval, yval;
         if (mystrtod(&p, &val)) {
-            if (render_priv->state.shadow_x == render_priv->state.shadow_y)
-                val = render_priv->state.shadow_x * (1 - pwr) + val * pwr;
+            xval = render_priv->state.shadow_x * (1 - pwr) + val * pwr;
+            yval = render_priv->state.shadow_y * (1 - pwr) + val * pwr;
+            // VSFilter compatibility: clip for \shad but not for \[xy]shad
+            xval = (xval < 0) ? 0 : xval;
+            yval = (yval < 0) ? 0 : yval;
         } else
-            val = 0.;
-        render_priv->state.shadow_x = render_priv->state.shadow_y = val;
+            xval = yval = render_priv->state.style->Shadow;
+        render_priv->state.shadow_x = xval;
+        render_priv->state.shadow_y = yval;
     } else if (mystrcmp(&p, "s")) {
         int val;
-        if (mystrtoi(&p, &val) && val)
+        if (!mystrtoi(&p, &val) || !(val == 0 || val == 1))
+            val = render_priv->state.style->StrikeOut;
+        if (val)
             render_priv->state.flags |= DECO_STRIKETHROUGH;
         else
             render_priv->state.flags &= ~DECO_STRIKETHROUGH;
     } else if (mystrcmp(&p, "u")) {
         int val;
-        if (mystrtoi(&p, &val) && val)
+        if (!mystrtoi(&p, &val) || !(val == 0 || val == 1))
+            val = render_priv->state.style->Underline;
+        if (val)
             render_priv->state.flags |= DECO_UNDERLINE;
         else
             render_priv->state.flags &= ~DECO_UNDERLINE;
     } else if (mystrcmp(&p, "pbo")) {
-        double val = 0;
-        if (mystrtod(&p, &val))
-            render_priv->state.drawing->pbo = val;
+        double val;
+        mystrtod(&p, &val);
+        render_priv->state.pbo = val;
     } else if (mystrcmp(&p, "p")) {
         int val;
-        if (!mystrtoi(&p, &val))
-            val = 0;
-        if (val)
-            render_priv->state.drawing->scale = val;
-        render_priv->state.drawing_mode = !!val;
+        mystrtoi(&p, &val);
+        val = (val < 0) ? 0 : val;
+        render_priv->state.drawing_scale = val;
     } else if (mystrcmp(&p, "q")) {
         int val;
-        if (!mystrtoi(&p, &val))
+        if (!mystrtoi(&p, &val) || !(val >= 0 && val <= 3))
             val = render_priv->track->WrapStyle;
         render_priv->state.wrap_style = val;
+    } else if (mystrcmp(&p, "fe")) {
+        int val;
+        if (!mystrtoi(&p, &val))
+            val = render_priv->state.style->Encoding;
+        render_priv->state.font_encoding = val;
     }
 
     return p;
@@ -890,7 +905,79 @@ void apply_transition_effects(ASS_Renderer *render_priv, ASS_Event *event)
 }
 
 /**
- * \brief Get next ucs4 char from string, parsing and executing style overrides
+ * \brief determine karaoke effects
+ * Karaoke effects cannot be calculated during parse stage (get_next_char()),
+ * so they are done in a separate step.
+ * Parse stage: when karaoke style override is found, its parameters are stored in the next glyph's
+ * (the first glyph of the karaoke word)'s effect_type and effect_timing.
+ * This function:
+ * 1. sets effect_type for all glyphs in the word (_karaoke_ word)
+ * 2. sets effect_timing for all glyphs to x coordinate of the border line between the left and right karaoke parts
+ * (left part is filled with PrimaryColour, right one - with SecondaryColour).
+ */
+void process_karaoke_effects(ASS_Renderer *render_priv)
+{
+    GlyphInfo *cur, *cur2;
+    GlyphInfo *s1, *e1;      // start and end of the current word
+    GlyphInfo *s2;           // start of the next word
+    int i;
+    int timing;                 // current timing
+    int tm_start, tm_end;       // timings at start and end of the current word
+    int tm_current;
+    double dt;
+    int x;
+    int x_start, x_end;
+
+    tm_current = render_priv->time - render_priv->state.event->Start;
+    timing = 0;
+    s1 = s2 = 0;
+    for (i = 0; i <= render_priv->text_info.length; ++i) {
+        cur = render_priv->text_info.glyphs + i;
+        if ((i == render_priv->text_info.length)
+            || (cur->effect_type != EF_NONE)) {
+            s1 = s2;
+            s2 = cur;
+            if (s1) {
+                e1 = s2 - 1;
+                tm_start = timing + s1->effect_skip_timing;
+                tm_end = tm_start + s1->effect_timing;
+                timing = tm_end;
+                x_start = 1000000;
+                x_end = -1000000;
+                for (cur2 = s1; cur2 <= e1; ++cur2) {
+                    x_start = FFMIN(x_start, d6_to_int(cur2->bbox.xMin + cur2->pos.x));
+                    x_end = FFMAX(x_end, d6_to_int(cur2->bbox.xMax + cur2->pos.x));
+                }
+
+                dt = (tm_current - tm_start);
+                if ((s1->effect_type == EF_KARAOKE)
+                    || (s1->effect_type == EF_KARAOKE_KO)) {
+                    if (dt >= 0)
+                        x = x_end + 1;
+                    else
+                        x = x_start;
+                } else if (s1->effect_type == EF_KARAOKE_KF) {
+                    dt /= (tm_end - tm_start);
+                    x = x_start + (x_end - x_start) * dt;
+                } else {
+                    ass_msg(render_priv->library, MSGL_ERR,
+                            "Unknown effect type");
+                    continue;
+                }
+
+                for (cur2 = s1; cur2 <= e1; ++cur2) {
+                    cur2->effect_type = s1->effect_type;
+                    cur2->effect_timing = x - d6_to_int(cur2->pos.x);
+                }
+                s1->effect = 1;
+            }
+        }
+    }
+}
+
+
+/**
+ * \brief Get next ucs4 char from string, parsing UTF-8 and escapes
  * \param str string pointer
  * \return ucs4 code of the next char
  * On return str points to the unparsed part of the string
@@ -899,24 +986,6 @@ unsigned get_next_char(ASS_Renderer *render_priv, char **str)
 {
     char *p = *str;
     unsigned chr;
-    if (*p == '{') {            // '\0' goes here
-        p++;
-        while (1) {
-            p = parse_tag(render_priv, p, 1.);
-            if (*p == '}') {    // end of tag
-                p++;
-                if (*p == '{') {
-                    p++;
-                    continue;
-                } else
-                    break;
-            } else if (*p != '\\')
-                ass_msg(render_priv->library, MSGL_V,
-                        "Unable to parse: '%.30s'", p);
-            if (*p == 0)
-                break;
-        }
-    }
     if (*p == '\t') {
         ++p;
         *str = p;
